{$G+,N+}                                                { 286+, 287+ }
{$R-,S-}

(*************************************************)
(* The Perfect Texture Mapping Algorithm Example *)
(*         !Numeric processor is requred!        *)
(*                Final release!                 *)
(*           All bugs have been fixed!           *)
(*                                               *)
(*        (c) 1999 by Alexei A. Frounze          *)
(*     Homepage: http://www.chat.ru/~alexfru     *)
(*           E-mail: alexfru@chat.ru             *)
(*************************************************)

Uses CRT, Grafix, GfxFont;


                            (* TYPE DEFINITIONS *)

Type
  TVector = Record                      { 3d point/vector definition }
    X, Y, Z : Single;
  End;


                               (* CONSTANTS *)

Const
  Screen_Width = 320;
  Screen_Height = 200;
  Center_X = Screen_Width div 2;
  Center_Y = Screen_Height div 2;
  Texture_Exp = 7;
  Texture_Size = 1 shl Texture_Exp;     { Should be a power of 2 }
  EL = Texture_Size*1;                  { Edge length equals texture size }
                                        { Feel free to put *2 or something }
                                        { else, now this works! }

  TMethod : Word = 3;                   { Texture mapping method (1...3) }

  MethodStr : Array [1..3] of String = (
    'Pas+floating point arithmetic in inner loop.',
    'TAsm+both arithmetics+8-pxl interpolation.',
    'TAsm+both arithmetics+16-pxl interpolation.'
  );

  CVertices : Array[0..7] of TVector =  { Standing cube vertices }
   ((X:-EL;Y:-EL;Z:-EL), (X:EL;Y:-EL;Z:-EL), (X:EL;Y:-EL;Z:EL), (X:-EL;Y:-EL;Z:EL),
    (X:-EL;Y: EL;Z:-EL), (X:EL;Y: EL;Z:-EL), (X:EL;Y: EL;Z:EL), (X:-EL;Y: EL;Z:EL));

  Polys : Array[0..5,0..3] of Word =    { Sets of sides' vertices }
    ((0,1,5,4), (1,2,6,5), (2,3,7,6), (3,0,4,7), (3,2,1,0), (4,5,6,7));


                               (* VARIABLES *)

Var
  PScreen : Pointer;                            { Virtual screen pointer }
  Texture : Array[0..Texture_Size-1,0..Texture_Size-1] of Byte; { Texture }
  Palette : TPal;                               { Color palette }

  Distance : Double;                            { Distance to the cube center }
  Vertices : Array[0..7] of TVector;            { Current verices }
  SinTab,                                       { Sine table }
  CosTab   : Array[0..359] of Double;           { Cosine table }

  Scan_Buf_L,
  Scan_Buf_R : Array [0..Screen_Height-1] of LongInt; { Scan buffer for texture mapper }
  XR, YR : Array[0..3] of Double;               { Polygon's coordinates }
  YminI, YmaxI : LongInt;                       { Polygon's screen min & max y coords }


{$L tline.obj}
procedure tline8 (VOFFS : Word; LLEN : Word; PTEXTURE : Pointer; L3, L1, L2, I3, I1, I2 : Comp); far; external;
procedure tline16 (VOFFS : Word; LLEN : Word; PTEXTURE : Pointer; L3, L1, L2, I3, I1, I2 : Comp); far; external;

                (* ADDITIONAL REAL NUMBER ROUNDING FUNCTIONS *)

{$L CEILFLOR.OBJ}
Function Floor (X : Double) : LongInt; Near; External;  { Rounds down }
Function Ceil (X : Double) : LongInt; Near; External;   { Rounds up }


                         (* VECTOR MATH FUNCTIONS *)

Procedure Vector_Add (Var A, B, C : TVector);
Begin
  C.X := A.X + B.X;
  C.Y := A.Y + B.Y;
  C.Z := A.Z + B.Z
End;

Procedure Vector_Sub (Var A, B, C : TVector);
Begin
  C.X := A.X - B.X;
  C.Y := A.Y - B.Y;
  C.Z := A.Z - B.Z
End;

Function Vector_Length (Var A : TVector) : Double;
Begin
  Vector_Length := Sqrt( Sqr(A.X) + Sqr(A.Y) + Sqr(A.Z) )
End;

Procedure Vector_Normalize (Var A : TVector);
Var L : Double;
Begin
  L := Vector_Length (A);
  A.X := A.X / L;
  A.Y := A.Y / L;
  A.Z := A.Z / L
End;

Function Dot_Product (Var A, B : TVector) : Double;
Begin
  Dot_Product := A.X*B.X + A.Y*B.Y + A.Z*B.Z
End;

Procedure Cross_Product (Var A, B, C : TVector);
Begin
  C.X := A.Y*B.Z - A.Z*B.Y;
  C.Y := A.Z*B.X - A.X*B.Z;
  C.Z := A.X*B.Y - A.Y*B.X
End;


                           (* INIT/DONE FUNCTIONS *)

Procedure Load_Texture;
Var F : File;
Begin
  Assign (F, 'TEXTURE.DAT');
  Reset (F, 1);
  BlockRead (F, Texture, Sqr(Texture_Size));
  Close (F);
  Assign (F, 'PALETTE.DAT');
  Reset (F, 1);
  BlockRead (F, Palette, 768);
  Close (F)
End;

Procedure Set_Up;
Var
  I : Integer;
  Alpha : Double;
Begin
  GetMem (PScreen, 64000);
  SetActivePage (Seg(PScreen^));
  For I := 0 to 359 do Begin
    Alpha := I*PI/180;
    SinTab[I] := Sin (Alpha);
    CosTab[I] := Cos (Alpha)
  End
End;

Procedure Done;
Begin
  SetActivePage (VSeg);
  FreeMem (PScreen, 64000)
End;


                       (* TEXTURE MAPPING FUNCTIONS *)

Procedure Scan_Edge (N : Word);
Var
  X, DX : Double;
  IsLeftEdge : Boolean;
  X1R, Y1R, X2R, Y2R : Double;
  Y1I, Y2I, YI, XI : LongInt;
  I, H : Word;
Begin

  { Getting both ends coordinates. }

  X1R := XR[N]; Y1R := YR[N];
  X2R := XR[(N+1) and 3]; Y2R := YR[(N+1) and 3];

  { Is line left or right? Determining that. }

  IsLeftEdge := Y2R < Y1R;              { Because of the clockwise order. }

  { Let's perpare beginning values of x and y. }

  If IsLeftEdge then Begin
    Y2I := Ceil(Y2R); Y1I := Ceil(Y1R)-1;
    If Y2I > Y1I then Exit;
    DX := (X1R - X2R) / (Y1R - Y2R);
    X := X2R + (Y2I - Y2R) * DX;
    YI := Y1I; Y1I := Y2I; Y2I := YI
  End
  Else Begin
    Y1I := Ceil(Y1R); Y2I := Ceil(Y2R)-1;
    If Y1I > Y2I then Exit;
    DX := (X2R - X1R) / (Y2R - Y1R);
    X := X1R + (Y1I - Y1R) * DX;
  End;

  { Do clipping! }

  If Y1I < 0 then Begin
    X := X - Y1I * DX;
    Y1I := 0
  End;
  If Y2I > Screen_Height-1 then Begin
    Y2I := Screen_Height-1
  End;

  { Scan through the line! }

  If IsLeftEdge then
    For YI := Y1I to Y2I do Begin
      XI := Ceil(X);
      Scan_Buf_L[YI] := XI;
      X := X + DX
    End
  Else
    For YI := Y1I to Y2I do Begin
      XI := Ceil(X) - 1;
      Scan_Buf_R[YI] := XI;
      X := X + DX
    End
End;

Procedure T_Map (Var P, M, N : TVector);
Var
  PM, NP, MN : TVector;
  IndexMin, IndexMax, I : Word;
  Y, XLI, XRI : LongInt;
  YminR, YmaxR,
  D, DDX, DDY,
  DU, DUDX, DUDY,
  DV, DVDX, DVDY,
  WD, WDU, WDV,
  IZ : Double;
  X, U, V : Word;
  C : Byte;
  PP : PChar;
  T1, T2, T3 : Double;
Label
  LSkip;
Begin

  { Let's find the coner points }

  XR[0] := P.X / P.Z;
  YR[0] := P.Y / P.Z;

  T1 := P.X+M.X; T2 := P.Y+M.Y; T3 := P.Z+M.Z;
  XR[1] := T1 / T3;
  YR[1] := T2 / T3;

  XR[2] := (T1+N.X) / (T3+N.Z);
  YR[2] := (T2+N.Y) / (T3+N.Z);

  XR[3] := (P.X+N.X) / (P.Z+N.Z);
  YR[3] := (P.Y+N.Y) / (P.Z+N.Z);

  { Let's transform the points to the screen coordinate system. }

  For I := 0 to 3 do Begin
    XR[I] := Center_X+256*XR[I];
    YR[I] := Center_Y+256*YR[I]
  End;

  { Let's find the top and the bottom points. }

  IndexMin := 0;
  IndexMax := 0;
  For I := 1 to 3 do Begin
    If YR[I] < YR[IndexMin] then IndexMin := I;
    If YR[I] > YR[IndexMax] then IndexMax := I
  End;

  YminR := YR[IndexMin];
  YmaxR := YR[IndexMax];
  YminI := Ceil(YminR);
  YmaxI := Ceil(YmaxR)-1;

  { Is the polygon visible at the screen? Skip if it isn't. }

  If YminI > YmaxI then Exit;
  If (YminI >= Screen_Height) or (YmaxI < 0) then Exit;
  If YminI < 0 then YminI := 0;
  If YmaxI >= Screen_Height then YmaxI := Screen_Height-1;

  { Do scan for the each edge. }

  For I := 0 to 3 do
    Scan_Edge (I);

  { And now we're preparing some vectors and constants }
  { to use in the inner loops. }

  Cross_Product (P, M, PM);
  Cross_Product (N, P, NP);
  Cross_Product (M, N, MN);

  DDX := MN.X;
  DDY := MN.Y;
  D := MN.Z*256 + (YminI-Center_Y)*DDY;

  DUDX := NP.X;
  DUDY := NP.Y;
  DU := NP.Z*256 + (YminI-Center_Y)*DUDY;

  DVDX := PM.X;
  DVDY := PM.Y;
  DV := PM.Z*256 + (YminI-Center_Y)*DVDY;

  { Going down through the polygon screen projection. }

  For Y := YminI to YmaxI do Begin

    { Getting left and right boundaries XLI and XRI. }

    XLI := Scan_Buf_L[Y]; If XLI < 0 then XLI := 0;
    XRI := Scan_Buf_R[Y]; If XRI > Screen_Width-1 then XRI := Screen_Width-1;

    WD := D + (XLI-Center_X)*DDX;
    WDU := DU + (XLI-Center_X)*DUDX;
    WDV := DV + (XLI-Center_X)*DVDX;

    { Go ahead! Texturate the horizontal line! }

    If XLI > XRI then Goto LSkip;

    PP := Ptr (VSegA, XLI);
    Inc (PP, (Y shl 6)+(Y shl 8));

    Case TMethod of
      1: { Pascal source + floating point. } Begin
          For X := XLI to XRI do Begin
            IZ := EL / WD;
            U := Trunc(WDU * IZ) and (Texture_Size-1);
            V := Trunc(WDV * IZ) and (Texture_Size-1);
            C := Texture[V,U];
            {PutPixel (X, Y, C);}
            PP^ := Char(C);
            Inc (PP);
            WD := WD + DDX;
            WDU := WDU + DUDX;
            WDV := WDV + DVDX
          End
        End;
      2: { TAsm source + mixed fixed and floating point + }
         { scan line subdivision for every 8 pixels. }
        TLine8 (Ofs(PP^), XRI-XLI+1, Addr(Texture),
               WD, WDU*EL, WDV*EL,
               DDX, DUDX*EL, DVDX*EL);
      3: { TAsm source + mixed fixed and floating point + }
         { scan line subdivision for every 16 pixels. }
        TLine16 (Ofs(PP^), XRI-XLI+1, Addr(Texture),
               WD, WDU*EL, WDV*EL,
               DDX, DUDX*EL, DVDX*EL)
    End;

LSkip:
    D := D + DDY;
    DU := DU + DUDY;
    DV := DV + DVDY
  End
End;


                        (* CUBE ROTATING FUNCTIONS *)

Procedure Rotate_Vector_X (No, Angle : Word);
Var
  V : TVector;
Begin
  V.Y := (Vertices[No].Y * CosTab[Angle]  +  Vertices[No].Z * SinTab[Angle]);
  V.Z := (Vertices[No].Z * CosTab[Angle]  -  Vertices[No].Y * SinTab[Angle]);
  V.X := Vertices[No].X;
  Vertices[No] := V
End;

Procedure Rotate_Vector_Y (No, Angle : Word);
Var
  V : TVector;
Begin
  V.Z := (Vertices[No].Z * CosTab[Angle]  +  Vertices[No].X * SinTab[Angle]);
  V.X := (Vertices[No].X * CosTab[Angle]  -  Vertices[No].Z * SinTab[Angle]);
  V.Y := Vertices[No].Y;
  Vertices[No] := V
End;

Procedure Rotate_Vector_Z (No, Angle : Word);
Var
  V : TVector;
Begin
  V.X := (Vertices[No].X * CosTab[Angle]  +  Vertices[No].Y * SinTab[Angle]);
  V.Y := (Vertices[No].Y * CosTab[Angle]  -  Vertices[No].X * SinTab[Angle]);
  V.Z := Vertices[No].Z;
  Vertices[No] := V
End;

Procedure Rotate_Cube (AngleX, AngleY, AngleZ : Word);
Var I : Word;
Begin
  For I := 0 to 7 do Begin
    Vertices[I] := CVertices[I];
    Rotate_Vector_X (I, AngleX);
    Rotate_Vector_Y (I, AngleY);
    Rotate_Vector_Z (I, AngleZ);
    Vertices[I].Z := Vertices[I].Z + Distance
  End
End;


                         (* CUBE DRAWING FUNCTIONS *)

Procedure Draw_Cube;
Var
  I, J, K : Word;
  P, M, N, Normal : TVector;
Begin
  { Drawing sides of the cube. }

  For I := 0 to 5 do Begin
    P := Vertices[Polys[I,0]];
    Vector_Sub (Vertices[Polys[I,1]], Vertices[Polys[I,0]], M);
    Vector_Sub (Vertices[Polys[I,3]], Vertices[Polys[I,0]], N);
    Cross_Product (M, N, Normal);
    If Normal.Z > 0 then                { Hidden face removing }
      T_Map (P, M, N)
  End
End;


                   (* ADDITIONAL FUNCTIONS AND VARIABLES *)

Function BIOS_Time : LongInt;
Begin
  BIOS_Time := MemL[Seg0040:$6C]
End;

Var
  XA, YA, ZA : Word;                    { Rotating angles }
  C : Char;
  Wait4Retrace : Boolean;               { Wait for vertical retrace flag }
  T0, T1, Frames : LongInt;             { Time }
  S, FS : String;
Label
  LEnd;

                              (* MAIN PROGRAM *)

Begin
  If Test8086 < 2 then Begin
    WriteLn ('80386 or better processor is required.');
    Goto LEnd
  End;
  If Test8087 < 3 then Begin
    WriteLn ('80387 or better numeric processor is required.');
    Goto LEnd
  End;
  If not IsVGA then Begin
    WriteLn ('VGA or SVGA adapeter is required.');
    Goto LEnd
  End;

  ClrScr;
  WriteLn ('-= The Perfect Texture Mapping Algorithm Example =-');
  WriteLn ('Have fun with this cool program! It shows different');
  WriteLn ('methods of perfect texture mapping.'#13#10);
  WriteLn ('Control keys:');
  WriteLn ('  1   - For the 1st method (slowest)');
  WriteLn ('  2   - For the 2nd method (good)');
  WriteLn ('  3   - For the 3rd method (fastest)');
  WriteLn ('  V   - Vertical retrace waiting switch');
  WriteLn ('  A   - Move the cube farther');
  WriteLn ('  Z   - Move the cube closer');
  WriteLn ('  Esc - Quit'#13#10);
  Write ('Press any key to continue...'); ReadKey;

  Load_Texture;
  Set_Up;
  InitGraph;
  SetFont (@Font6x7);
  SetPal (Palette);
  SetCRGB (0, 0, 0, 0);
  SetCRGB (255, 15, 63, 15);

  Distance := 5*EL;                     { Distance to the center of cube }
  XA := 0; YA := 0; ZA := 0;
  Wait4Retrace := True;

  T0 := BIOS_Time;
  T1 := T0;
  Frames := 0;
  Str (Frames:3, FS);

  Repeat
    If KeyPressed then C := UpCase(ReadKey)
    Else C := #$FF;
    If C = #27 then Break;
    If (C = 'Z') and (Distance > 2*EL) then Distance := Distance - EL/2;
    If (C = 'A') and (Distance < 20*EL) then Distance := Distance + EL/2;
    If C = 'V' then Wait4Retrace := not Wait4Retrace;
    If C = '1' then TMethod := 1;
    If C = '2' then TMethod := 2;
    If C = '3' then TMethod := 3;

    Rotate_Cube (XA, YA, ZA);           { Rotate the cube }
    ClearScreen (Black);                { Clear virtual sreen }
    Draw_Cube;                          { Draw the cube }

    WriteS (0, 0, BLACK, 255, 'Method:'+MethodStr[TMethod]);
    If Wait4Retrace then S := 'On' Else S := 'Off';
    WriteS (0, 7, BLACK, 255, 'Retrace waiting:'+S);

    If T1 - T0 >= 18 then Begin
      Str (Frames:3, FS);
      T0 := T1;
      Frames := 0;
    End;
    WriteS (0, 14, BLACK, 255, 'FPS:'+FS);

    If Wait4Retrace then WaitRetrace;   { Wait for retrace }
    DisplayPage (VSegA);                { Flip the virtual screen }
    XA := (XA+1) mod 360;               { Update }
    YA := (YA+3) mod 360;                {  the   }
    ZA := (ZA+5) mod 360;                 { angles }
    T1 := BIOS_Time;
    Inc (Frames)
  Until False;

  CloseGraph;
  Done;
LEnd:
  WriteLn ('Ŀ');
  WriteLn (' The Perfect Texture Mapping Algorithm Example ');
  WriteLn ('        (c) 1999 by Alexei A. Frounze          ');
  WriteLn ('     Homepage: http://www.chat.ru/~alexfru     ');
  WriteLn ('            E-mail: alexfru@chat.ru            ');
  WriteLn ('');
End.
